#include "stdafx.h"
#include "resource.h"
#include <windows.h>
#include <stdlib.h>
#include <math.h>
#include "vector.h"

#define WIDTH		512
#define HEIGHT		384
#define SIZE		(WIDTH * HEIGHT * 4)

const char appName[] = " ";
const char className[] = "RayTracing";
const double M_PI = 3.1415926535897932384626433832795;
const double DEG = 180 / M_PI;
const double GeomThreshold = 0.001;

HANDLE RenderThread;
unsigned char *render_buf;
HDC hDC;
HWND hWnd;
BITMAPINFOHEADER bi;

struct Medium {
	double refr;
	double betta;
	Medium() { }
	Medium(double r, double b)
	{
		refr = r;
		betta = b;
	}
};

struct SurfaceData {
	double Ka;
	double Kd;
	double Ks;
	double Kr;
	double Kt;
	double p;
	Medium med;
};

struct Ray {
	Vector org, dir;
	Ray() { }
	Ray(Vector o, Vector d)
	{
		org = o;
		dir = d;
    }
	Ray(Ray& ray)
	{
		org = ray.org;
		dir = ray.dir;
	}
	Vector Point(double dist)
	{
		return org + dir * dist;
	}
};

struct Light {
	Vector color, org;
	Light *next;
	Light(Vector o, Vector c)
	{
		next = NULL;
		org = o;
		color = c;
	}
	virtual Vector GetColor(double, Vector, Vector) = 0;
};

struct PointLight: public Light {
	PointLight(Vector org, Vector color): Light(org, color) { }
	Vector GetColor(double density, Vector dir, Vector n)
	{
		return color * exp(-density * (!dir)) * fabs(Normalize(dir) & n);
	}
};

class SpotLight: public PointLight {
	Vector dir;
	double angle, blur;
public:
	SpotLight(Vector org, Vector color, Vector d, double a, double b):
		PointLight(org, color), dir(d), angle(a)
	{
		blur = b;
	}
	Vector GetColor(double density, Vector d, Vector n)
	{
		double a = acos((dir & -d) / (!dir * !d)) * DEG;
		if (a < angle)
		{
			Vector c = PointLight::GetColor(density, d, n);
			if (a < angle - blur) return c;
			else return c * (1 - (a - angle + blur) / blur);
		}
		return 0;
	}
};

struct GObject {
	SurfaceData data;
    GObject *next;
	GObject() { }
	GObject(SurfaceData d)
    {
		data = d;
		next = NULL;
	}
	virtual bool Intersect(Ray ray, double& dist) = 0;
	virtual Vector GetColor(Vector point) = 0;
	virtual Vector GetNormal(Vector point) = 0;
};

class Plane: public GObject {
    double d;
    int i, j, k;
    Vector c;
public:
    Plane(SurfaceData data, double d_, int i_, Vector color): GObject(data)
    {
		d = d_;
		i = i_;
		j = k = 0;
		c = color;
    }
    bool Intersect(Ray ray, double& dist)
    {
		if ((ray.dir[i] >= 0.0 && ray.org[i] > d) || (ray.dir[i] <= 0.0 && ray.org[i] < d))
			return false;
		dist = fabs((ray.org[i] - d) / ray.dir[i]);
		return true;
    }
    Vector GetNormal(Vector point)
    {
		Vector res(0);
		res[i] = 1.0;
		return res;
    }
    Vector GetColor(Vector p)
    {
		return c;
    }
};

struct XYPlane: public Plane {
    XYPlane(SurfaceData data, double d, Vector color):
    Plane(data, d, 2, color) { }
};

struct YZPlane: public Plane {
    YZPlane(SurfaceData data, double d, Vector color):
    Plane(data, d, 0, color) { }
};

struct ZXPlane: public Plane {
    ZXPlane(SurfaceData data, double d, Vector color):
    Plane(data, d, 1, color) { }
};

class Sphere: public GObject {
    double r, r2;
    Vector org, color;
public:
    Sphere(SurfaceData data, Vector o, double r_, Vector c): GObject(data)
    {
		org = o;
		r = r_;
		r2 = r * r;
		color = c;
    }
    bool Intersect(Ray ray, double& dist)
    {
		Vector l = org - ray.org;
		double l2oc = l & l;
		double tca = l & ray.dir;
		double t2hc = r2 - l2oc + tca * tca;
		double t2;

		if (t2hc < 0) return false;
		
		t2hc = sqrt(t2hc);
		if (tca < t2hc)
		{
			dist = tca + t2hc;
			t2 = tca - t2hc;
		}
		else
		{
			dist = tca - t2hc;
			t2 = tca + t2hc;
		}

		if (fabs(dist) < GeomThreshold) dist = t2;

		return dist > GeomThreshold;
    }
    Vector GetNormal(Vector p)
    {
		return (p - org) / r;
    }
    Vector GetColor(Vector p)
    {
		return color;
    }
};

class Scene {
    Medium med, old_med;
    int cur_ray, max_rays;
    long width, height;
    double last_dist;
    char *buf;
    Vector bgcolor, ambient;
    GObject *beg_obj, *end_obj;
    Light *beg_light, *end_light;
    Vector Trace(Ray, GObject *, Medium, GObject *);
    GObject *Intersect(Ray, GObject *, double&);
public:
    Scene(Vector, Vector, Medium);
    ~Scene();
    void AddObject(GObject *);
    void AddLight(Light *);
    void RenderScene(int, int, double, double, Vector, Vector, Vector, int, unsigned char *);
} *scene;

Scene::Scene(Vector bgcolor, Vector ambient, Medium med)
{
	cur_ray = 0;
	beg_obj = end_obj = NULL;
	beg_light = end_light = NULL;
	Scene::bgcolor = bgcolor;
	Scene::ambient = ambient;
	Scene::med = med;
	old_med = med;
}

Scene::~Scene()
{
	GObject *obj = beg_obj;
	Light *light = beg_light;

	while (obj != NULL)
	{
		beg_obj = obj->next;
		delete obj;
		obj = beg_obj;
	}

	while (light != NULL)
	{
		beg_light = light->next;
		delete light;
		light = beg_light;
	}
}

void Scene::AddObject(GObject *obj)
{
	if (beg_obj == NULL) beg_obj = obj;
	else end_obj->next = obj;
	end_obj = obj;
}

void Scene::AddLight(Light *light)
{
	if (beg_light == NULL) beg_light = light;
	else end_light->next = light;
	end_light = light;
}

void Scene::RenderScene(int width, int height, double nx, double ny, Vector camera_org, Vector camera_dir, Vector up_dir, int maxrays, unsigned char *buf)
{
	max_rays = maxrays;
	Vector color, vx, vy;
	double x, y, x_step = 2.0 * nx / width, y_step = 2.0 * ny / height;
	Ray ray;
	int i, j;

	vx = Normalize(up_dir ^ camera_dir);
	vy = Normalize(camera_dir ^ vx);
	ray.org = camera_org;

	for (i = 0, y = ny; i < height; i++, y -= y_step)
	{
		for (j = 0, x = nx; j < width; j++, x -= x_step)
		{	
			ray.dir = Normalize(camera_dir + vx * x + vy * y);
			color = Trace(ray, NULL, med, NULL);
			color = Clip(color);
			*(buf++) = unsigned char(color.z * 255);
			*(buf++) = unsigned char(color.y * 255);
			*(buf++) = unsigned char(color.x * 255);
			*(buf++) = 0;
		}
		InvalidateRect(hWnd, NULL, FALSE);
	}
}

Vector Scene::Trace(Ray ray, GObject *ignore, Medium cur_med, GObject *cur_obj)
{
	Medium new_med;
	Vector color(bgcolor), reflect, refract, sum(0), ldir, point, n, v, sum2(0), c;
	GObject *o, *obj;
	Light *light = beg_light;
	double dist, t;
	Ray r;
	SurfaceData data;
	cur_ray++;

	if ((obj = Intersect(ray, ignore, dist)) != NULL)
	{
		point = ray.Point(dist);
		data = obj->data;
		n = obj->GetNormal(point);
		r.org = point;

		if (cur_ray < max_rays && data.Kr > 0 && cur_obj != obj)
		{
			r.dir = ray.dir - n * (2 * (ray.dir & n));
			reflect = Trace(r, obj, cur_med, cur_obj);
		}
		else reflect = 0;

		if (cur_ray < max_rays && data.Kt > 0)
		{
			double c;
			if (cur_obj == obj)
			{
				c = ray.dir & n;
				new_med = old_med;
			}
			else
			{
				c = -ray.dir & n;
				new_med = data.med;
			}
			old_med = cur_med;
			double et = cur_med.refr / new_med.refr;
			r.dir = et * ray.dir + (et * c - sqrt(1 + et * et * (c * c - 1))) * n;
			refract = Trace(r, NULL, new_med, obj);
		}
		else refract = 0;

		if (data.Kd > 0 || data.Ks > 0)
		{
			while (light != NULL)
			{
				v = light->org - point;
				ldir = Normalize(v);
				r.dir = ldir;
				o = Intersect(r, obj, t);
				if (o == NULL || t > !v)
				{
					c = light->GetColor(cur_med.betta, v, n);
					sum += c;
					sum2 += c * fabs(pow(n & Normalize(ldir - ray.dir), data.p));
				}
				light = light->next;
			}
		}

		color = data.Ka * ambient +
                data.Kd * obj->GetColor(point) * sum +
                data.Ks * sum2 +
                data.Kr * reflect * exp(-cur_med.betta * last_dist) +
                data.Kt * refract * exp(-cur_med.betta * last_dist);
	}
	else
	{
		color = bgcolor;
		dist = 1.7E+308;
	}

	cur_ray--;
	last_dist = dist;
	return color;
}

GObject *Scene::Intersect(Ray ray, GObject *ignore, double& dist)
{
	GObject *obj = beg_obj, *o = NULL;
	double min = 1.7E+308;

	while (obj != NULL)
	{
		if (obj != ignore)
			if (obj->Intersect(ray, dist))
				if (dist < min)
				{
					min = dist;
					o = obj;
				}
		obj = obj->next;
	}

	dist = min;
	return o;
}

DWORD WINAPI ThreadFunc(LPVOID)
{
	scene->RenderScene(WIDTH, HEIGHT, 0.64, 0.48, Vector(0, 0, 2), Vector(1, 0, 0), Vector(0, 0, 1), 4, render_buf);
	return 0;
}

void CreateScene()
{
	SurfaceData data;
	data.Ka = 0;
	data.Kd = 1;
	data.Kr = 0;
	data.Ks = 0;
	data.Kt = 0;
	data.p = 0;
	data.med = Medium(1.02, 0.05);
	scene = new Scene(Vector(0.8, 0.8, 1.0), Vector(0), Medium(1, 0.05));
	scene->AddObject(new XYPlane(data, 0, Vector(0.5, 0.5, 0.5)));
	data.Kd = 0.3;
	data.Ks = 0.3;
	data.Kr = 0.4;
	data.p = 4;
	scene->AddObject(new Sphere(data, Vector(10, -2, 2), 2, Vector(1, 1, 0)));
	data.Kd = 0.2;
	data.Ks = 0.5;
	data.Kr = 0.2;
	data.Kt = 0.5;
	scene->AddObject(new Sphere(data, Vector(7, 3, 1), 1, Vector(0, 0, 1)));
	data.Kt = 0;
	data.Kd = 1;
	data.Ks = 0;
	data.Kr = 0;
	scene->AddObject(new Sphere(data, Vector(14, 3, 1.5), 1.5, Vector(1, 0, 0)));
	data.Kd = 0;
	data.Ks = 0;
	data.Kr = 1;
	scene->AddObject(new Sphere(data, Vector(6, -3, 0.5), 0.5, Vector(1)));
	data.Kd = 0.8;
	data.Kr = 0.2;
	data.Ks = 0;
	data.p = 0;
	scene->AddObject(new XYPlane(data, 6, Vector(1, 1, 1)));
	data.Kd = 1;
	data.Kr = 0;
	scene->AddObject(new YZPlane(data, 18, Vector(0, 1, 0)));
	scene->AddObject(new YZPlane(data, -2, Vector(0, 1, 0)));
	scene->AddObject(new ZXPlane(data, 6, Vector(1, 1, 0)));
	scene->AddObject(new ZXPlane(data, -6, Vector(1, 1, 0)));
	scene->AddLight(new PointLight(Vector(15, 3, 5), Vector(1)));
	scene->AddLight(new SpotLight(Vector(6, -3, 2), Vector(1, 1, 1), Vector(0, 0, -1), 45, 10));
	scene->AddLight(new PointLight(Vector(0, 0, 3), Vector(0.7, 0.7, 1)));
}

LONG WINAPI wndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	PAINTSTRUCT ps;
	HBITMAP Bitmap;
	HDC MemDC;
	RECT r;
	DWORD id;

	switch(msg)
	{
		case WM_CREATE:
			hDC = GetDC(hWnd);
			CreateScene();
			return TRUE;

		case WM_PAINT:
			BeginPaint(hWnd, &ps);
			MemDC = CreateCompatibleDC(hDC);
			Bitmap = CreateCompatibleBitmap(hDC, WIDTH, HEIGHT);
			SetDIBits(hDC, Bitmap, 0, HEIGHT, render_buf, (BITMAPINFO *)&bi, DIB_RGB_COLORS);
			SelectObject(MemDC, Bitmap);
			BitBlt(hDC, 0, 0, WIDTH, HEIGHT, MemDC, 0, 0, SRCCOPY);
			DeleteDC(MemDC);
			DeleteObject(Bitmap);
			EndPaint(hWnd, &ps);
			return TRUE;

		case WM_SIZING:
			GetWindowRect(hWnd, &r);
			memcpy((void *)lParam, &r, sizeof(RECT));
			return TRUE;

		case WM_COMMAND:
			switch (LOWORD(wParam))
			{
				case ID_RENDER:
					ZeroMemory(render_buf, SIZE);
					RenderThread = CreateThread(NULL, 0, ThreadFunc, NULL, 0, &id);
					return TRUE;

				case ID_EXITAPP:
					DestroyWindow(hWnd);
					return TRUE;
			}
			break;

		case WM_DESTROY:
			TerminateThread(RenderThread, 0);
			CloseHandle(RenderThread);
			ReleaseDC(hWnd, hDC);
			delete scene;
			PostQuitMessage(0);
			return TRUE;

		default:
			return DefWindowProc(hWnd, msg, wParam, lParam);
	}

	return FALSE;
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR cmdLine, int cmdShow)
{
	WNDCLASS wc;
	MSG msg;
	RECT r, r2;
	int w1, h1, h2, w2;

	bi.biBitCount = 32;
	bi.biClrImportant = 0;
	bi.biClrUsed = 0;
	bi.biCompression = BI_RGB;
	bi.biHeight = -HEIGHT;
	bi.biPlanes = 1;
	bi.biSize = sizeof(bi);
	bi.biSizeImage = WIDTH * HEIGHT * 4;
	bi.biWidth = WIDTH;
	bi.biXPelsPerMeter = 3780;
	bi.biYPelsPerMeter = 3780;

	render_buf = new unsigned char[SIZE];
	ZeroMemory(render_buf, SIZE);

	wc.style = CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc =(WNDPROC)wndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hIcon = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_APP));
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground =(HBRUSH)GetStockObject(BLACK_BRUSH);
	wc.lpszMenuName = MAKEINTRESOURCE(IDR_MENU);
	wc.lpszClassName = className;

	RegisterClass(&wc);

	hWnd = CreateWindow(className, appName, WS_OVERLAPPEDWINDOW,
						CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, 
						HWND_DESKTOP, NULL, hInstance, NULL);

	ShowWindow(hWnd, cmdShow);
	UpdateWindow(hWnd);

	GetWindowRect(hWnd, &r);
	GetClientRect(hWnd, &r2);
	w1 = r.right - r.left;
	h1 = r.bottom - r.top;
	w2 = r2.right - r2.left;
	h2 = r2.bottom - r2.top;
	r.right = r.left + WIDTH + (w1 - w2);
	r.bottom = r.top + HEIGHT + (h1 - h2);
	SetWindowPos(hWnd, 0, r.left, r.top, r.right - r.left, r.bottom - r.top, 0);

	while(GetMessage(&msg, NULL, 0, 0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	delete render_buf;
	return msg.wParam;
}

